/*-------------------------------------------------------------------*/
/* CBERROR.C      Juha Kivijrvi                                     */
/*                Timo Kaukoranta                                    */
/*                                                                   */
/* - Calculates average error between codebook and training set.     */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "CBERROR"
#define VersionNumber  "V.0.23"
#define LastUpdated    "2.11.98"

/* ----------------------------------------------------------------- */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>

#define  FACTFILE  "cberror.fac"

#include "parametr.c"
#include "cb.h"
#include "file.h"
#include "interfc.h"
#include "memctrl.h"
#include "owntypes.h"

#define width(a)   ( (a) > 0  ? (int) log10(a) + 1 : \
                     (a) == 0 ? 1 : (int) log10(-a) +2 )

/*============================  P R I N T  ===============================*/


static void PrintOperatingInfo(char  TSetName[],
                               char  CBName[])
{
  if( Value(ShowProgress) >= 1 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
    printf("Training Set: %s\n", TSetName);
    printf("Codebook:     %s\n", CBName);
    PrintSelectedOptions();
    printf("\n");
    }
  else
    {
    printf("%-12s ", TSetName);
    printf("%-12s ", CBName);
    }
}


/*-------------------------------------------------------------------*/


void PrintInfo(void)
{
  printf("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
  printf("This program calculates statistics of a training set and a codebook.\n");
  printf("Usage: %s [%coption] <training-set> <codebook>\n", ProgName, OPTION_SYMBOL);
  printf("For example: %s set_of_4 genetic\n", ProgName);
  printf("\n  Options:\n");
  PrintOptions();
  printf("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
}


/*=========================  M A T H  ======================================*/


static double Average(double* values, int* freqs, int count)
{
  int    i;
  double sum     = 0.0;
  int    freqsum = 0;

  if (count == 0) return 0.0;
  
  for (i=0; i<count; i++)
    {
    sum     += values[i] * freqs[i];
    freqsum += freqs[i];
    }
  
  return (sum / freqsum);
}
  

/*-------------------------------------------------------------------*/


static double StandardDeviation(double* values, 
                                int*    freqs,
                                int     count, 
                                double  average)
{
  int    i;
  double diffsum = 0.0;
  int    freqsum = 0;
  
  if (count == 0) return 0.0;

  for (i=0; i<count; i++)
    {
    diffsum += (values[i] - average) * (values[i] - average) * freqs[i];
    freqsum += freqs[i];
    }
    
  return sqrt(diffsum / freqsum);
}    


/*-------------------------------------------------------------------*/


static double Dissimilarity(TRAININGSET* TS, PARTITIONING* P, int index)
{
  int     i,j;
  double* errors;
  int*    freqs;
  int     count;
  int     ind = 0;
  double  aver;
  double  stdev;

  if (UniqueVectors(P,index) < 2) return 0.0;

  count  = (UniqueVectors(P,index) * (UniqueVectors(P,index) + 1)) / 2;

  errors = allocate( count * sizeof(double) );
  freqs  = allocate( count * sizeof(int) );

  for (i=FirstVector(P,index); !(EndOfPartition(i)); i=NextVector(P,i))
    {
    errors[ind] = 0;
    freqs[ind]  = VectorFreq(TS,i) * (VectorFreq(TS,i)-1) / 2;
    ind++;
    for (j=FirstVector(P,index); i!=j; j=NextVector(P,j))
      {
      errors[ind] = sqrt( (double) VectorDist(Vector(TS,i),
                                              Vector(TS,j),
                                              VectorSize(TS)));
      freqs[ind]  = VectorFreq(TS,i) * VectorFreq(TS,j);
      ind++;
      }
    }

  aver  = Average(errors, freqs, count);
  stdev = StandardDeviation(errors, freqs, count, aver);

  deallocate(errors);
  deallocate(freqs);
  return ( aver < 0.000000001 ? 0.0 : stdev / aver );  
}      


/*-------------------------------------------------------------------*/


static void CalculateMAEandNMatches(TRAININGSET* TS, 
                                    CODEBOOK* CB,
                                    PARTITIONING* P,
                                    double* MAE,
                                    int* nMatches)
{
  int    i,j;
  int    MAEdiff;
  llong  MAEsum    = 0LL;

  *nMatches = 0;
  for (i=0; i<BookSize(TS); i++)
    {
    MAEdiff = 0;
    for (j=0; j<VectorSize(TS); j++)
      {
      MAEdiff += VectorFreq(TS,i) * abs( VectorScalar(TS,i,j) - 
                                         VectorScalar(CB,Map(P,i),j) );
      }
    if (MAEdiff == 0) (*nMatches)++;
    MAEsum += MAEdiff;
    }
  *MAE = (double) MAEsum / ((double)TotalFreq(TS) * (double)VectorSize(TS));
}


/*=========================  CODEBOOK HANDLING  ============================*/


static void CodeTSWithCB(TRAININGSET*  TS, 
                         CODEBOOK*     CB, 
                         PARTITIONING* P,
                         double        MSE)
{
  int    i,j;
  double MAE = 0.0;
  int    nMatches  = 0;
  llong  PErrorSum = 0LL;
  int    vectorlength;
  double dissim;
  double dissimsum = 0.0;

  CalculateMAEandNMatches(TS, CB, P, &MAE, &nMatches);

  if( Value(ShowProgress) >= 1 )
    {
    printf("MSE Distortion = %9.4f\n", PrintableError(MSE,CB));
    printf("MAE Distortion = %9.4f\n", PrintableError(MAE,CB));
    printf("Equal vectors  = %5d\n",   nMatches);
    }
  else
    {
    printf("%9.4f ", PrintableError(MSE,CB));
    printf("%9.4f ", PrintableError(MAE,CB));
    printf("%5d\n",  nMatches);
    }
  if( Value(PartitionErrors) )
    {
    printf("\nind   ");
    vectorlength = VectorSize(CB) * (width(CB->MaxValue) + 
                   (Value(NoBlanks) ? 0 : 1)) + 1;

    if (vectorlength < 11)
      for (i=0; i<vectorlength; i++) printf(" ");
    else
      {
      printf("codevector");
      for (i=10; i<vectorlength; i++) printf(" ");
      }
    printf(" %5s %5s %8s %9s %9s\n", 
           "freq", "uniq", "sqerror", "MSE", "dissim");

    for (i=0; i<BookSize(CB); i++)
      {
      PErrorSum = 0LL;
      for (j=FirstVector(P,i); !(EndOfPartition(j)); j=NextVector(P,j))
        PErrorSum += VectorDist( Vector(TS,j), Vector(CB,i), VectorSize(TS) ) *
                     VectorFreq(TS,j);
        
      printf("%3d | ", i+1);

      for (j=0; j<VectorSize(CB); j++)
        {
        printf("%*d", width(CB->MaxValue), VectorScalar(CB,i,j));
        if (!(Value(NoBlanks))) printf(" ");
        }                         
      if (Value(NoBlanks)) printf(" ");
      printf("| %5d %5d %8d ",CCFreq(P,i), UniqueVectors(P,i), (int)PErrorSum);

      printf("%9.4f ", ( CCFreq(P,i) == 0 ? 0.0 : (double)PErrorSum /                                                  
                        ((double)CCFreq(P,i) * VectorSize(CB)) ) ); 
      dissim = Dissimilarity(TS,P,i);
      dissimsum += dissim * CCFreq(P, i);;
      printf("%9.4f\n", dissim);
             
      }
    printf("Average dissimilarity = %f\n", dissimsum / TotalFreq(TS));

    }
}


/* ----------------------------------------------------------------- */


static void CodeBothWays(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* P)
{
  double MSE1 = 0.0, MSE2 = 0.0;
  double MAE1 = 0.0, MAE2 = 0.0;
  int    nMatches1 = 0, nMatches2 = 0;

  MSE1 = GenerateOptimalPartitioning(TS, CB, P);

  CalculateMAEandNMatches(TS, CB, P, &MAE1, &nMatches1);

  MSE2 = GenerateOptimalPartitioning(CB, TS, P);

  CalculateMAEandNMatches(CB, TS, P, &MAE2, &nMatches2);

  if( Value(ShowProgress) >= 1 )
    {
    printf("                  TS->CB    CB->TS    Average\n");
    printf("MSE Distortion = %9.4f %9.4f %9.4f\n", 
            PrintableError(MSE1,CB), 
            PrintableError(MSE2,CB),
            PrintableError((MSE1+MSE2)/2.0,CB));
    printf("MAE Distortion = %9.4f %9.4f %9.4f\n", 
            PrintableError(MAE1,CB), 
            PrintableError(MAE2,CB),
            PrintableError((MAE1+MAE2)/2.0,CB));
    printf("Equal vectors  = %9d %9d\n", nMatches1, nMatches2);
    }
  else
    {
    printf("%9.4f %9.4f %9.4f\n", PrintableError(MSE1,CB),             
                                  PrintableError(MSE2,CB),             
                                  PrintableError((MSE1+MSE2)/2.0,CB)); 
            
    }
}


/* ----------------------------------------------------------------- */


static void DoCalculations(TRAININGSET*  TS, 
                           CODEBOOK*     CB, 
                           PARTITIONING* P,
                           CBFILETYPE    filetype)
{
  double     MSEv;
  ERRORFTYPE errorf;

  if (Value(ErrorFunction) == MSE_D)
    {
    if ( filetype == CBFILE || filetype == TSFILE )
      {
      if( Value(DoubleWayError) == NO )
        {
        MSEv = GenerateOptimalPartitioning(TS, CB, P);
        CodeTSWithCB(TS, CB, P, MSEv);
        }
      else
        {
        if ( BookSize(TS) == BookSize(CB) )
          CodeBothWays(TS, CB, P);
        else
          printf("ERROR: TrainingSet and Codebook sizes differ.\n");
        }
      }
    else
      {
      GenerateOptimalCodebook(TS, CB, P);
      MSEv = AverageErrorForSolution(TS, CB, P, MSE);
      CodeTSWithCB(TS, CB, P, MSEv);
      }
    }
  else 
    { 
    switch (Value(ErrorFunction))
        {
        case SC_D:      errorf = SC;          break;
        case JACCARD_D: errorf = JACCARD;     break;
        case DICE_D:    errorf = DICE;        break;
        default:        errorf = MSE;
        }
    if ( filetype == CBFILE || filetype == TSFILE )
      {
      MSEv = AverageErrorCB(TS, CB, errorf);
      }
    else
      {
      MSEv = AverageErrorForPartitioning(TS, P, errorf);
      }
    printf("Distortion = %9.4f\n", MSEv);
    }

}


/* ================================================================= */


static void Initialize(TRAININGSET*  TS, 
                       CODEBOOK*     CB, 
                       PARTITIONING* P,
                       CBFILETYPE    filetype)
{
  int ndups;

  if( (ndups = DuplicatesInCodebook(TS)) > 0 )
    {
    printf("\nWARNING: TrainingSet contains %i duplicates.\n\n", ndups);
    }

  if (filetype == CBFILE || filetype == TSFILE)
    {
    CreateNewPartitioning(P, TS, BookSize(CB));
    if( (ndups = DuplicatesInCodebook(CB)) > 0 )
      {
      printf("\nWARNING: Codebook contains %i duplicates.\n\n", ndups);
      }
    }
  else
    {
    CreateNewCodebook(CB, P->PartitionCount, TS);
    }
}


/* ================================================================= */


static void ShutDown(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* P)
{
  FreeCodebook(TS);
  FreeCodebook(CB);
  FreePartitioning(P);
}


/*===========================  M A I N  ==============================*/


int main(int argc, char* argv[])
{
  CODEBOOK      CB;
  TRAININGSET   TS;
  PARTITIONING  P;
  char          TSName[80];
  char          CBName[80];
  CBFILETYPE    ft1, ft2;

  ParameterInfo paraminfo[2] =  { { TSName, "", 0, INFILE  }, 
                                  { CBName, "", 0, INFILE  } };
                               
  ParseParameters( argc, argv, 2, paraminfo);
  ft1 = DetermineCBFileTypeConsideringOrder(TSName, 12);
  ft2 = DetermineCBFileTypeConsideringOrder(CBName, 231);
  if ( (ft1==CBFILE && ft2==PAFILE) || ft1==NOTFOUND || ft2==NOTFOUND )
    {
    PrintInfo();
    exit(-1);
    }

  PrintOperatingInfo(TSName, CBName);
  ReadTrainingSet(TSName, &TS);
  if(ft2 == PAFILE) ReadPartitioning(CBName, &P, &TS);
  else              ReadCodebook(CBName, &CB);
  Initialize(&TS, &CB, &P, ft2);

  DoCalculations(&TS, &CB, &P, ft2);

  ShutDown(&TS, &CB, &P);
  checkmemory();
  return( 0 );
}

